#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implementation of the zero-shot classifier."""

import json
import logging
import os
from typing import Any, List, Optional, Union

import torch
from paccmann_predictor.models import MODEL_FACTORY
from pytoda.proteins.protein_language import ProteinLanguage
from pytoda.smiles.smiles_language import SMILESLanguage
from pytoda.transforms import LeftPadding, ToTensor

from ....frameworks.torch import device_claim

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class MCAPredictor:
    """Base implementation of an MCAPredictor."""

    def predict(self) -> Any:
        """Get prediction.

        Returns:
            predicted affinity
        """
        raise NotImplementedError("No prediction implemented for base MCAPredictor")

    def predict_values(self) -> Any:
        """Get prediction for algorithm sample method.

        Returns:
            predicted values as list.
        """
        raise NotImplementedError(
            "No values prediction implemented for base MCAPredictor"
        )


class BimodalMCAAffinityPredictor(MCAPredictor):
    """Bimodal MCA (Multiscale Convolutional Attention) affinity prediction model.

    For details see: https://pubs.acs.org/doi/10.1021/acs.molpharmaceut.9b00520
    and https://iopscience.iop.org/article/10.1088/2632-2153/abe808.
    """

    def __init__(
        self,
        resources_path: str,
        protein_targets: List[str],
        ligands: List[str],
        confidence: bool,
        device: Optional[Union[torch.device, str]] = None,
    ):
        """Initialize BimodalMCAAffinityPredictor.

        Args:
            resources_path: path where to load model weights and cofiguration.
            protein_targets: list of protein targets as AA sequences.
            ligands: list of ligands in SMILES format.
            confidence: whether the confidence for the prediction should be returned.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        self.device = device_claim(device)
        self.resources_path = resources_path
        self.protein_targets = protein_targets
        self.ligands = ligands
        self.confidence = confidence

        # setting affinity predictor parameters
        with open(os.path.join(resources_path, "mca_model_params.json")) as f:
            self.predictor_params = json.load(f)
        self.affinity_predictor = MODEL_FACTORY["bimodal_mca"](self.predictor_params)
        self.affinity_predictor.load(
            os.path.join(resources_path, "mca_weights.pt"),
            map_location=self.device,
        )
        affinity_protein_language = ProteinLanguage.load(
            os.path.join(resources_path, "protein_language.pkl")
        )
        affinity_smiles_language = SMILESLanguage.load(
            os.path.join(resources_path, "smiles_language.pkl")
        )
        self.affinity_predictor._associate_language(affinity_smiles_language)
        self.affinity_predictor._associate_language(affinity_protein_language)
        self.affinity_predictor.eval()

        self.pad_smiles_predictor = LeftPadding(
            self.affinity_predictor.smiles_padding_length,
            self.affinity_predictor.smiles_language.padding_index,
        )

        self.pad_protein_predictor = LeftPadding(
            self.affinity_predictor.protein_padding_length,
            self.affinity_predictor.protein_language.padding_index,
        )

        self.to_tensor = ToTensor()

    def predict(self) -> Any:
        """Get predicted affinity.

        Returns:
            predicted affinity.
        """
        # prepare ligand representation
        ligand_tensor = torch.cat(
            [
                torch.unsqueeze(
                    self.to_tensor(
                        self.pad_smiles_predictor(
                            self.affinity_predictor.smiles_language.smiles_to_token_indexes(
                                ligand_smiles
                            )
                        )
                    ),
                    0,
                )
                for ligand_smiles in self.ligands
            ],
            dim=0,
        )

        # prepare target protein representation
        target_tensor = torch.cat(
            [
                torch.unsqueeze(
                    self.to_tensor(
                        self.pad_protein_predictor(
                            self.affinity_predictor.protein_language.sequence_to_token_indexes(
                                protein_target
                            )
                        )
                    ),
                    0,
                )
                for protein_target in self.protein_targets
            ],
            dim=0,
        )

        with torch.no_grad():
            predictions, predictions_dict = self.affinity_predictor(
                ligand_tensor,
                target_tensor,
                confidence=self.confidence,
            )

        return predictions, predictions_dict

    def predict_values(self) -> List[float]:
        """Get prediction for algorithm sample method.

        Returns:
            predicted values as list.
        """
        predictions, _ = self.predict()
        return list(predictions[:, 0])



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""MoLeR Algorithm.

MoLeR generation algorithm.
"""

import logging
from dataclasses import field
from typing import Any, ClassVar, Dict, Optional, TypeVar

from ....domains.materials import SMILES, validate_molecules
from ....exceptions import InvalidItem
from ...core import AlgorithmConfiguration, GeneratorAlgorithm, Untargeted
from ...registry import ApplicationsRegistry
from .implementation import MoLeRGenerator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = type(None)
S = TypeVar("S", bound=SMILES)


class MoLeR(GeneratorAlgorithm[S, T]):
    """MoLeR Algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T] = None,
    ):
        """Instantiate MoLeR ready to generate items.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for generating small molecules (SMILES) with the default configuration:

                configuration = MoLeRDefaultGenerator()
                MoLeR = MoLeR(configuration=configuration, target=target)
                items = list(MoLeR.sample(10))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Untargeted:
        """Get the function to sample batches via the MoLeRGenerator.

        Args:
            configuration: helps to set up the application.
            target: context or condition for the generation. Unused in the algorithm.

        Returns:
            callable generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: MoLeRGenerator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.generate

    def validate_configuration(
        self, configuration: AlgorithmConfiguration[S, T]
    ) -> AlgorithmConfiguration[S, T]:
        # TODO raise InvalidAlgorithmConfiguration
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration


@ApplicationsRegistry.register_algorithm_application(MoLeR)
class MoLeRDefaultGenerator(AlgorithmConfiguration[SMILES, Any]):
    """Configuration to generate compounds using default parameters of MoLeR."""

    algorithm_type: ClassVar[str] = "generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    scaffolds: str = field(
        default="",
        metadata=dict(
            description="Scaffolds as '.'-separated SMILES. If empty, no scaffolds are used."
        ),
    )
    num_samples: int = field(
        default=32,
        metadata=dict(description="Number of molecules to sample per call."),
    )
    beam_size: int = field(
        default=1,
        metadata=dict(description="Beam size to use during decoding."),
    )
    seed: int = field(
        default=0,
        metadata=dict(description="Seed used for random number generation."),
    )
    num_workers: int = field(
        default=6,
        metadata=dict(description="Number of workers used for generation."),
    )

    def get_target_description(self) -> Optional[Dict[str, str]]:
        """Get description of the target for generation.

        Returns:
            target description, returns None in case no target is used.
        """
        return None

    def get_conditional_generator(self, resources_path: str) -> MoLeRGenerator:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate<gt4sd.algorithms.generation.MoLeR.implementation.MoLeRGenerator.generate>` for generation.
        """
        return MoLeRGenerator(
            resources_path=resources_path,
            scaffolds=self.scaffolds,
            num_samples=self.num_samples,
            beam_size=self.beam_size,
            seed=self.seed,
            num_workers=self.num_workers,
        )

    def validate_item(self, item: str) -> SMILES:
        """Check that item is a valid SMILES.

        Args:
            item: a generated item that is possibly not valid.

        Raises:
            InvalidItem: in case the item can not be validated.

        Returns:
            the validated SMILES.
        """
        (
            molecules,
            _,
        ) = validate_molecules([item])
        if molecules[0] is None:
            raise InvalidItem(
                title="InvalidSMILES",
                detail=f'rdkit.Chem.MolFromSmiles returned None for "{item}"',
            )
        return SMILES(item)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
from typing import Generator, Union

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import IterableDataset

from ..envs.graph_building_env import GraphBuildingEnvContext
from ..loss.trajectory_balance import TrajectoryBalance, TrajectoryBalanceModel
from .dataset import GFlowNetDataset, GFlowNetTask

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class SamplingIterator(IterableDataset):
    """Sampler to speed up training based on IterableDataset."""

    def __init__(
        self,
        dataset: GFlowNetDataset,
        model: Union[nn.Module, TrajectoryBalanceModel],
        batch_size: int,
        context: GraphBuildingEnvContext,
        algorithm: TrajectoryBalance,
        task: GFlowNetTask,
        ratio: float = 0.5,
        stream: bool = True,
        device: str = "cuda",
    ) -> None:
        """
        This class allows us to parallelise and train faster.
        By separating sampling data/the model and building torch geometric
        graphs from training the model, we can do the former in different
        processes, which is much faster since much of graph construction
        is CPU-bound. This sampler can handle offline and online data.

        Code adapted from: https://github.com/recursionpharma/gflownet/tree/trunk/src/gflownet/data.

        Args:
            dataset: a dataset instance.
            model: the model we sample from (must be on CUDA already or share_memory() must be called so that
                parameters are synchronized between each worker).
            batch_size: the number of trajectories, each trajectory will be comprised of many graphs, so this is
                _not_ the batch size in terms of the number of graphs (that will depend on the task).
            context: the graph environment.
            algorithm: the training algorithm, e.g. a TrajectoryBalance instance.
            task: ConditionalTask that specifies the reward structure.
            ratio: the ratio of offline trajectories in the batch.
            stream: if true, data is sampled iid for every batch. Otherwise, this is a normal in-order
                dataset iterator.
            device: the device (cpu, cuda).
        """
        self.data = dataset
        self.model = model
        self.batch_size = batch_size
        self.offline_batch_size = int(np.ceil(batch_size * ratio))
        self.online_batch_size = int(np.floor(batch_size * (1 - ratio)))
        self.ratio = ratio
        self.ctx = context
        self.algo = algorithm
        self.task = task
        self.device = device
        self.stream = stream

    def _idx_iterator(self) -> Generator:
        """Returns an iterator over the indices of the dataset. The batch can be offline or online.

        Yields:
            Batch of indexes.
        """
        bs = self.offline_batch_size

        n = len(self.data)

        if self.stream:
            # if we're streaming data, just sample `offline_batch_size` indices
            while True:
                yield self.rng.integers(0, n, bs)  # type: ignore
        else:
            # figure out which indices correspond to this worker
            worker_info = torch.utils.data.get_worker_info()

            # refactor this
            if worker_info is None:
                start = 0
                end = n
                wid = -1
            else:
                nw = worker_info.num_workers
                wid = worker_info.id
                start = int(np.floor(n / nw * wid))
                end = int(np.ceil(n / nw * (wid + 1)))

            if end - start < bs:
                yield np.arange(start, end)
                return
            for i in range(start, end - bs, bs):
                yield np.arange(i, i + bs)
            if i + bs < end:
                yield np.arange(i + bs, end)

    def __len__(self):
        """Length of the offline or online dataset.

        Returns:
            The length of the dataset.
        """
        # if online
        if self.stream:
            return int(1e6)
        # if offline
        return len(self.data)  # type: ignore

    def sample_offline(self, idcs):
        """Samples offline data.

        Args:
            idcs: the indices of the data to sample.

        Returns:
            trajs: the trajectories.
            rewards: the rewards.
        """
        # sample offline batch (mols, rewards)
        mols, _flat_rewards = map(list, zip(*[self.data[i] for i in idcs]))
        # rewards
        flat_rewards = list(self.task.flat_reward_transform(_flat_rewards))  # type: ignore
        # build graphs
        graphs = [self.ctx.mol_to_graph(m) for m in mols]
        # use trajectory balance to sample trajectories
        trajs = self.algo.create_training_data_from_graphs(graphs)
        return trajs, flat_rewards

    def predict_reward_model(self, trajs, flat_rewards, num_offline):
        """Predict rewards using the model.

        Args:
            trajs: the trajectories.
            flat_rewards: the rewards.
            num_offline: the number of offline trajectories.

        Returns:
            flat_rewards: the updated rewards.
        """
        # The model can be trained to predict its own reward,
        # i.e. predict the output of cond_info_to_reward
        pred_reward = [i["reward_pred"].cpu().item() for i in trajs[num_offline:]]
        flat_rewards += list(pred_reward)
        return flat_rewards

    def predict_reward_task(self, trajs, flat_rewards, num_offline, is_valid):
        """Predict rewards using the task.

        Args:
            trajs: the trajectories.
            flat_rewards: the rewards.
            num_offline: the number of offline trajectories.
            is_valid: whether the trajectories are valid.

        Returns:
            flat_rewards: the updated rewards.
        """
        # Otherwise, query the task for flat rewards
        valid_idcs = torch.tensor(
            [
                i + num_offline
                for i in range(self.online_batch_size)
                if trajs[i + num_offline]["is_valid"]
            ]
        ).long()

        pred_reward = torch.zeros((self.online_batch_size))
        # fetch the valid trajectories endpoints
        mols = [self.ctx.graph_to_mol(trajs[i]["traj"][-1][0]) for i in valid_idcs]

        # ask the task to compute their reward
        preds, m_is_valid = self.task.compute_flat_rewards(mols)
        # The task may decide some of the mols are invalid, we have to again filter those
        valid_idcs = valid_idcs[m_is_valid]
        _preds = torch.tensor(preds, dtype=torch.float32)
        pred_reward[valid_idcs - num_offline] = _preds

        is_valid[num_offline:] = False
        is_valid[valid_idcs] = True
        flat_rewards += list(pred_reward)

        # Override the is_valid key in case the task made some mols invalid
        for i in range(self.online_batch_size):
            trajs[num_offline + i]["is_valid"] = is_valid[num_offline + i].item()

        return trajs, flat_rewards

    def sample_online(self, trajs, flat_rewards, cond_info, num_offline):
        """Sample on-policy data.

        Args:
            trajs: the trajectories.
            flat_rewards: the rewards.
            cond_info: the conditional information.
            num_offline: the number of offline trajectories.

        Returns:
            trajs: the updated trajectories.
            flat_rewards: the updated rewards.
        """
        is_valid = torch.ones(cond_info["beta"].shape[0]).bool()

        with torch.no_grad():
            trajs += self.algo.create_training_data_from_own_samples(
                self.model,  # TODO: double-check the model here
                self.online_batch_size,
                cond_info["encoding"][num_offline:],
            )

        # predict reward with model
        if self.algo.bootstrap_own_reward:
            flat_rewards = self.predict_reward_model(trajs, flat_rewards, num_offline)

        # predict reward with task
        else:
            trajs, flat_rewards = self.predict_reward_task(
                trajs, flat_rewards, num_offline, is_valid
            )

        return trajs, flat_rewards

    def set_seed(self):
        """Set the seed for the workers."""
        worker_info = torch.utils.data.get_worker_info()
        wid = worker_info.id if worker_info is not None else 0
        # set seed for each worker
        seed = np.random.default_rng(142857 + wid)
        self.rng = seed
        self.algo.rng = seed
        self.task.rng = seed
        self.ctx.device = self.device

    def __iter__(self):
        """Build batch using online and offline data and multiple workers.

        Yields:
            batch of data using trajectories and rewards.
        """
        # we need to set a different seed for each worker to sample different batches.
        # If we start with the same seed, each worker has the exact same copy of the data and will yield the same batch.
        self.set_seed()

        # import ipdb
        # ipdb.sset_trace()
        # iterate over the indices in the batch and yield a bunch of indexes
        for idcs in self._idx_iterator():
            num_offline = idcs.shape[0]  # This is in [1, self.offline_batch_size]
            # Sample conditional info such as temperature, trade-off weights, etc.
            cond_info = self.task.sample_conditional_information(
                num_offline + self.online_batch_size
            )
            is_valid = torch.ones(cond_info["beta"].shape[0]).bool()

            # sample offline data
            trajs, flat_rewards = self.sample_offline(idcs)

            # Sample some on-policy data (sample online the model or the task)
            if self.online_batch_size > 0:
                # update trajectories and rewards with on-policy data
                trajs, flat_rewards = self.sample_online(
                    trajs, flat_rewards, cond_info, num_offline
                )

            # compute scalar rewards from conditional information & flat rewards
            rewards = self.task.cond_info_to_reward(cond_info, flat_rewards)
            # account for illegal actions
            rewards[torch.logical_not(is_valid)] = np.exp(
                self.algo.illegal_action_logreward
            )

            # Construct batch using trajectories, rewards
            batch = self.algo.construct_batch(trajs, cond_info["encoding"], rewards)
            batch.num_offline = num_offline
            yield batch



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import List, Union

from rdkit import Chem

from .core import PropertyPredictor
from .molecules import AVAILABLE_MOLECULES_PROPERTY_PREDICTOR
from .proteins import AVAILABLE_PROTEINS_PROPERTY_PREDICTOR
from .scores import DistanceScorer


class PropertyPredictorScorer(DistanceScorer):
    """Property Predictor Scorer."""

    def __init__(
        self,
        name: str,
        scoring_function: PropertyPredictor,
        target: Union[float, int],
    ) -> None:
        """Scoring function that calculates a generic score for a property.

        Args:
            name: name of the property to score. Needed for validation.
            scoring_function: callable scoring function.
            target: target score that will be used to get the distance to the score of a molecule or protein (not be confused with parameters["target"]).
        """
        self.name = name
        self.target = target

        self.scoring_function = scoring_function
        super().__init__()

    def score(self, sample: str) -> float:
        """Generates a score for a given molecule or protein.

        Args:
            sample: molecule or protein.

        Returns:
            A score for the given molecule or protein.
        """

        self.validate_input(sample=sample, property_name=self.name)
        return self.get_distance(self.scoring_function(sample) - self.target)

    def score_list(self, sample_list: List[str]) -> List[float]:
        """Generates a list of scores for a given molecule or protein list.

        Args:
            samples_list: A List of molecules or proteins.

        Returns:
            A List of scores
        """
        scores = []
        for sample in sample_list:
            scores.append(self.score(sample))
        return scores

    def predictor(self, sample: str) -> Union[float, int]:
        """Generates a prediction for a given molecule or protein.

        Args:
            sample: molecule or protein.

        Returns:
            A score for the given SMILES
        """

        self.validate_input(sample=sample, property_name=self.name)
        return self.scoring_function(sample)

    def validate_input(self, sample: str, property_name: str) -> None:
        """Validates the sample in input.
        If self.name is a property available for molecules, check that sample is a SMILES.
        If self.name is a property available for proteins, check that sample is a protein.

        Args:
            sample: molecule or protein.

        Raises:
            ValueError: if the sample is not a valid SMILES or protein given a certain property.

        Returns:
            True if the input is valid.
        """
        # if selected property is available for molecules
        if self.name in AVAILABLE_MOLECULES_PROPERTY_PREDICTOR:
            # check that sample is a valid SMILES, if not raise error
            if Chem.MolFromSmiles(sample) is None:
                raise ValueError(
                    f"{property_name} is a property available for molecules and {sample} is not a valid SMILES. Please input a molecule."
                )
        # if selected property is available for proteins
        elif self.name in AVAILABLE_PROTEINS_PROPERTY_PREDICTOR:
            # check that sample is a valid FASTA, if not raise error
            if Chem.MolFromFASTA(sample) is None:
                raise ValueError(
                    f"{property_name} is a property available for proteins and {sample} is not a valid FASTA. Plese input a protein."
                )
        # if property is not available, raise error
        else:
            raise ValueError(
                f"{property_name} is not available or not a valid property."
            )


class MoleculePropertyPredictorScorer(PropertyPredictorScorer):
    """Property Predictor Scorer for molecules."""

    def __init__(
        self,
        name: str,
        scoring_function: PropertyPredictor,
        target: Union[float, int],
    ) -> None:
        """Scoring function that calculates a generic score for a property in molecules.

        Args:
            name: name of the property to score.
            scoring_function: callable scoring function.
            target: target score that will be used to get the distance to the score of a molecule or protein (not be confused with parameters["target"]).
        """
        if name not in AVAILABLE_MOLECULES_PROPERTY_PREDICTOR:
            raise ValueError(f"property {name} not available for molecules.")

        super().__init__(name=name, scoring_function=scoring_function, target=target)


class ProteinPropertyPredictorScorer(PropertyPredictorScorer):
    """Property Predictor Scorer for protein."""

    def __init__(
        self,
        name: str,
        scoring_function: PropertyPredictor,
        target: Union[float, int],
    ) -> None:
        """Scoring function that calculates a generic score for a property in proteins.

        Args:
            name: name of the property to score.
            scoring_function: callable scoring function.
            target: target score that will be used to get the distance to the score of a molecule or protein (not be confused with parameters["target"]).
        """
        if name not in AVAILABLE_PROTEINS_PROPERTY_PREDICTOR:
            raise ValueError(f"property {name} not available for proteins.")

        super().__init__(name=name, scoring_function=scoring_function, target=target)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implementation of Reinvent conditional generators."""

import logging
import os
from typing import List, NamedTuple, Optional, Set, Tuple

from reinvent_models.lib_invent.models.model import DecoratorModel

from .reinvent_core.core import ReinventBase, SampledSequencesDTO

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class SampledTuple(NamedTuple):
    scaffold: str
    decoration: str
    nll: float


class ReinventConditionalGenerator(ReinventBase):
    def __init__(
        self,
        resources_path: str,
        batch_size: int,
        randomize: bool,
        sample_uniquely: bool,
        max_sequence_length: int,
    ):
        """Initialize Reinvent.

        Args:
            resources_path: path where to load hypothesis, candidate labels and, optionally, the model.
            batch_size: number of samples to generate per scaffold.
            randomize: randomize the scaffolds if set to true.
            sample_uniquely: generate unique sample sequences if set to true.
            max_sequence_length: maximum length of the generated sequences.
        """
        self.resources_path = resources_path
        self.batch_size = batch_size
        self.randomize = randomize
        self.sample_uniquely = sample_uniquely
        self.model_path = os.path.join(self.resources_path, "model.prior")
        self.target: Optional[str] = None

        if not os.path.isfile(self.model_path):
            logger.debug("reinvent model files does not exist locally")
            raise OSError(f"artifacts file {self.model_path} does not exist locally")

        self.model = DecoratorModel.load_from_file(path=self.model_path)
        self.model.max_sequence_length = max_sequence_length
        super().__init__(
            self.model, self.batch_size, self.randomize, self.sample_uniquely
        )

    def sample_unique_sequences(self, sampled_sequences: List[Tuple]) -> List[Tuple]:
        """
        Samples the model for the given number of SMILES.

        Args:
            scaffold_list: A list of SampledTuple.
        Returns:
            A list of SampledTuple.
        """
        sequences = [
            SampledSequencesDTO(scaffold, decoration, nll)
            for scaffold, decoration, nll in sampled_sequences
        ]
        logger.info("getting unique sample sequences from generated samples")
        return [
            (sample.scaffold, sample.decoration, sample.nll)
            for sample in self._sample_unique_sequences(sequences)
        ]

    def generate_sampled_tuples(self, scaffold: str) -> Set[SampledTuple]:
        """
        Samples the model for the given number of SMILES.
        Args:
            scaffold_list: A list of scaffold SMILES.
        Returns:
            A Set of SampledTuple.
        """
        if self.target != scaffold:
            self.target = scaffold
            batch = next(iter(self.get_dataloader([scaffold])))
            logger.info("initialization of the dataloader")
            scaffold_seqs, scaffold_seq_lengths = batch
            self.scaffold_seqs = scaffold_seqs.expand(
                self.batch_size - 1, scaffold_seqs.shape[1]
            )
            self.scaffold_seq_lengths = scaffold_seq_lengths.expand(self.batch_size - 1)
        logger.info("started generating samples with an nll score value")
        sampled_sequences = list(
            self.model.sample_decorations(self.scaffold_seqs, self.scaffold_seq_lengths)
        )
        if self.sample_uniquely:
            sampled_sequences = self.sample_unique_sequences(sampled_sequences)

        return set(
            [
                SampledTuple(scaffold, decoration, nll)
                for scaffold, decoration, nll in sampled_sequences
            ]
        )

    def generate_samples(self, scaffold: str) -> Set[str]:
        """
        Samples the model for the given number of SMILES.

        Args:
            scaffold: A scaffold SMILES.
        Returns:
            A Set of SMILES representing molecules.
        """
        return set(
            molecule for _, molecule, _ in self.generate_sampled_tuples(scaffold)
        )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Regression Transformer training utilities."""
import os
from dataclasses import dataclass, field
from typing import Optional

from ...configuration import gt4sd_configuration_instance
from ..core import TrainingPipelineArguments
from .utils import TransformersTrainingArgumentsCLI

DATA_ROOT = os.path.join(
    gt4sd_configuration_instance.gt4sd_local_cache_path, "data", "RegressionTransformer"
)
os.makedirs(DATA_ROOT, exist_ok=True)


@dataclass
class RegressionTransformerTrainingArguments(
    TrainingPipelineArguments, TransformersTrainingArgumentsCLI
):
    """
    Arguments related to RegressionTransformer trainer.
    NOTE: All arguments from `transformers.training_args.TrainingArguments` can be used.
    Only additional ones are specified below.
    """

    __name__ = "training_args"

    training_name: str = field(
        default="rt_training", metadata={"help": "Name used to identify the training."}
    )
    epochs: int = field(default=10, metadata={"help": "Number of epochs."})
    batch_size: int = field(default=16, metadata={"help": "Size of the batch."})
    log_interval: int = field(
        default=100, metadata={"help": "Number of steps between log intervals."}
    )
    gradient_interval: int = field(
        default=1, metadata={"help": "Gradient accumulation steps"}
    )

    max_span_length: int = field(
        default=5, metadata={"help": "Max length of a span of masked tokens for PLM."}
    )
    plm_probability: float = field(
        default=1 / 6,
        metadata={
            "help": "Ratio of length of a span of masked tokens to surrounding context length for PLM."
        },
    )
    alternate_steps: int = field(
        default=50,
        metadata={
            "help": "Per default, training alternates between property prediction and "
            "conditional generation. This argument specifies the alternation frequency."
            "If you set it to 0, no alternation occurs and we fall back to vanilla "
            "permutation language modeling (PLM). Default: 50."
        },
    )
    cc_loss: bool = field(
        default=True,
        metadata={
            "help": "Whether the cycle-consistency loss is computed during the conditional "
            "generation task. Defaults to True."
        },
    )


@dataclass
class RegressionTransformerDataArguments(TrainingPipelineArguments):
    """Arguments related to RegressionTransformer data loading."""

    __name__ = "dataset_args"

    train_data_path: str = field(
        metadata={
            "help": "Path to a `.csv` file with the input training data. The file has to "
            "contain a `text` column (with the string input, e.g, SMILES, AAS, natural "
            "text) and an arbitrary number of numerical columns."
        },
    )
    test_data_path: str = field(
        metadata={
            "help": "Path to a `.csv` file with the input testing data. The file has to "
            "contain a `text` column (with the string input, e.g, SMILES, AAS, natural "
            "text) and an arbitrary number of numerical columns."
        },
    )
    augment: Optional[int] = field(
        default=0,
        metadata={
            "help": "Factor by which the training data is augmented. The data modality "
            "(SMILES, SELFIES, AAS, natural text) is inferred from the tokenizer. "
            "NOTE: For natural text, no augmentation is supported. Defaults to 0, "
            "meaning no augmentation. "
        },
    )

    line_by_line: Optional[bool] = field(
        default=True,
        metadata={
            "help": "Whether lines of text in the dataset are to be handled as distinct samples."
        },
    )


@dataclass
class RegressionTransformerSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to RegressionTransformer trainer."""

    __name__ = "saving_args"

    model_path: str = field(
        metadata={"help": "Path where the model artifacts are stored."}
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Exceptions tests."""

import pytest

from gt4sd.exceptions import InvalidAlgorithmConfiguration, InvalidItem, S3SyncError


def test_s3_sync_error():
    error = S3SyncError("GenericSyncError", "my message")
    assert error.type == "S3SyncError"
    assert error.title == "GenericSyncError"
    assert error.detail == "my message"
    with pytest.raises(RuntimeError):
        str(error) == error.detail
        raise error


def test_invalid_item():
    error = InvalidItem("GenericInvaliItemError", "my message")
    assert error.type == "InvalidItem"
    assert error.title == "GenericInvaliItemError"
    assert error.detail == "my message"
    with pytest.raises(ValueError):
        str(error) == error.detail
        raise error


def test_invalid_algorithm_configuration():
    error = InvalidAlgorithmConfiguration(
        "GenericAlgorithmConfigurationError", "my message"
    )
    assert error.type == "InvalidAlgorithmConfiguration"
    assert error.title == "GenericAlgorithmConfigurationError"
    assert error.detail == "my message"
    with pytest.raises(ValueError):
        str(error) == error.detail
        raise error



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import queue
import threading

import torch
import torch.multiprocessing as mp
import torch.nn as nn


class MPModelPlaceholder:
    """This class can be used as a Model in a worker process, and
    translates calls to queries to the main process.
    """

    def __init__(self, in_queues, out_queues):
        self.qs = in_queues, out_queues
        self.device = torch.device("cpu")
        self._is_init = False

    def _check_init(self):
        if self._is_init:
            return
        info = torch.utils.data.get_worker_info()
        self.in_queue = self.qs[0][info.id]
        self.out_queue = self.qs[1][info.id]
        self._is_init = True

    def log_z(self, *a):
        self._check_init()
        self.in_queue.put(("log_z", *a))
        return self.out_queue.get()

    def __call__(self, *a):
        self._check_init()
        self.in_queue.put(("__call__", *a))
        return self.out_queue.get()


class MPModelProxy:
    """This class maintains a reference to an in-cuda-memory model, and
    creates a `placeholder` attribute which can be safely passed to
    multiprocessing DataLoader workers.

    This placeholder model sends messages accross multiprocessing
    queues, which are received by this proxy instance, which calls the
    model and sends the return value back to the worker.

    Starts its own (daemon) thread. Always passes CPU tensors between
    processes.
    """

    def __init__(self, model: torch.nn.Module, num_workers: int, cast_types: tuple):
        """Construct a multiprocessing model proxy for torch DataLoaders.

        Args:
            model: a torch model which lives in the main process to which method calls are passed.
            num_workers: number of workers.
            cast_types: types that will be cast to cuda when received as arguments of method calls.
        """
        self.in_queues = [mp.Queue() for i in range(num_workers)]  # type: ignore
        self.out_queues = [mp.Queue() for i in range(num_workers)]  # type: ignore
        self.placeholder = MPModelPlaceholder(self.in_queues, self.out_queues)
        self.model = model
        self.device = next(model.parameters()).device
        self.cuda_types = (torch.Tensor,) + cast_types
        self.stop = threading.Event()
        self.thread = threading.Thread(target=self.run, daemon=True)
        self.thread.start()

    def __del__(self):
        self.stop.set()

    def run(self):
        while not self.stop.is_set():
            for qi, q in enumerate(self.in_queues):
                try:
                    r = q.get(True, 1e-5)
                except queue.Empty:
                    continue
                except ConnectionError:
                    break
                attr, *args = r
                f = getattr(self.model, attr)
                args = [
                    i.to(self.device) if isinstance(i, self.cuda_types) else i
                    for i in args
                ]
                result = f(*args)
                if isinstance(result, (list, tuple)):
                    msg = [
                        i.detach().to(torch.device("cpu"))
                        if isinstance(i, self.cuda_types)
                        else i
                        for i in result
                    ]
                    self.out_queues[qi].put(msg)
                else:
                    msg = (
                        result.detach().to(torch.device("cpu"))
                        if isinstance(result, self.cuda_types)
                        else result
                    )
                    self.out_queues[qi].put(msg)


def wrap_model_mp(
    model: nn.Module, num_workers: int, cast_types: tuple
) -> MPModelPlaceholder:
    """Construct a multiprocessing model proxy for torch DataLoaders so
    that only one process ends up making cuda calls and holding cuda
    tensors in memory.

    Args:
        model: a torch model which lives in the main process to which method calls are passed.
        num_workers: number of DataLoader workers.
        cast_types: types that will be cast to cuda when received as arguments of method calls.
            torch.Tensor is cast by default.

    Returns:
        placeholder: a placeholder model whose method calls route arguments to the main process.

    """
    return MPModelProxy(model, num_workers, cast_types).placeholder



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Collection of available methods."""


import logging
from dataclasses import dataclass as vanilla_dataclass
from dataclasses import field, make_dataclass
from functools import WRAPPER_ASSIGNMENTS, update_wrapper
from typing import Any, Callable, ClassVar, Dict, List, NamedTuple, Optional, Type

import pydantic

# pyright (pylance in VSCode) does not support pydantic typechecking
# if typing.TYPE_CHECKING:
#     from dataclasses import dataclass
# else:
#     from pydantic.dataclasses import dataclass
from pydantic.dataclasses import dataclass

from ..exceptions import DuplicateApplicationRegistration
from .core import AlgorithmConfiguration, GeneratorAlgorithm

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ConfigurationTuple(NamedTuple):
    """Attributes to uniquely describe an AlgorithmConfiguration."""

    algorithm_type: str
    domain: str
    algorithm_name: str
    algorithm_application: str


class AnnotationTuple(NamedTuple):
    annotation: type
    default_value: Any  # TODO serializable type?


@vanilla_dataclass
class AlgorithmApplication:
    """Collect all needed to run an application."""

    algorithm_class: Type[GeneratorAlgorithm]
    configuration_class: Type[AlgorithmConfiguration]
    parameters_dict: Dict[str, AnnotationTuple] = field(default_factory=dict)
    # includes algorithm_version: str


class RegistryDict(Dict[ConfigurationTuple, AlgorithmApplication]):
    """Dict that raises when reassigning an existing key."""

    def __setitem__(self, key, value):
        if self.__contains__(key):
            raise DuplicateApplicationRegistration(
                title="Applications exists",
                detail=f"key {key} was already registered and would override another application.",
            )
            # if it's really needed for some reason, delete the item first, then add it.
        else:
            super().__setitem__(key, value)


class ApplicationsRegistry:
    """Registry to collect "applications" and make them accessible.

    An application denotes the combination of an
    :class:`AlgorithmConfiguration<gt4sd.algorithms.core.AlgorithmConfiguration>` and a
    :class:`GeneratorAlgorithm<gt4sd.algorithms.core.GeneratorAlgorithm>`.
    """

    # NOTE on import of registy also ensure import of modules to populate applications
    applications: RegistryDict = RegistryDict()

    @classmethod
    def _register_application(
        cls,
        algorithm_class: Type[GeneratorAlgorithm],
        algorithm_configuration_class: Type[AlgorithmConfiguration],
    ):
        # testing that configuration class is callable without arguments
        try:
            algorithm_configuration_class()
        except pydantic.ValidationError as e:
            logger.exception(e)
        config_tuple = cls.configuration_class_as_tuple(algorithm_configuration_class)
        cls.applications[config_tuple] = AlgorithmApplication(
            algorithm_class=algorithm_class,
            configuration_class=algorithm_configuration_class,
        )

    @classmethod
    def register_algorithm_application(
        cls,
        algorithm_class: Type[GeneratorAlgorithm],
        as_algorithm_application: Optional[str] = None,
    ) -> Callable[[Type[AlgorithmConfiguration]], Type[AlgorithmConfiguration]]:
        """Complete and register a configuration via decoration.

        Args:
            algorithm_class: The algorithm that uses the configuration.
            as_algorithm_application: Optional application name to use instead of
                the configurations class name.

        Returns:
            A function to complete the configuration class' attributes to reflect
            the matching GeneratorAlgorithm and application. The final class is
            registered and returned.

        Example:
            as decorator::

                from gt4sd.algorithms.registry import ApplicationsRegistry


                @ApplicationsRegistry.register_algorithm_application(SomeAlgorithm)
                class SomeApplication(AlgorithmConfiguration):
                    algorithm_type: ClassVar[str] = 'conditional_generation'
                    domain: ClassVar[str] = 'materials'
                    algorithm_version: str = 'v0'

                    some_more_serializable_arguments_with_defaults: int = 42

        Example:
            directly, here for an additional algorithm application with the same
            algorithm::

                AnotherApplication = ApplicationsRegistry.register_algorithm_application(
                    SomeAlgorithm, 'AnotherApplication'
                )(SomeApplication)
        """

        def decorator(
            configuration_class: Type[AlgorithmConfiguration],
        ) -> Type[AlgorithmConfiguration]:
            """Complete the configuration class' attributes and register the class.

            Args:
                configuration_class: class to complete.

            Returns:
                a completed class.
            """
            VanillaConfiguration = make_dataclass(
                cls_name=configuration_class.__name__,
                # call `@dataclass` for users to avoid confusion
                bases=(vanilla_dataclass(configuration_class),),
                fields=[
                    (
                        "algorithm_name",  # type: ignore
                        ClassVar[str],
                        field(default=algorithm_class.__name__),  # type: ignore
                    ),
                    (
                        "algorithm_application",  # type: ignore
                        ClassVar[str],
                        field(
                            default=(
                                as_algorithm_application or configuration_class.__name__  # type: ignore
                            )
                        ),
                    ),
                ],  # type: ignore
            )

            PydanticConfiguration: Type[AlgorithmConfiguration] = dataclass(  # type: ignore
                VanillaConfiguration
            )
            # get missing entries
            missing_in__dict__ = [
                key
                for key in configuration_class.__dict__
                if key not in PydanticConfiguration.__dict__
            ]

            update_wrapper(
                wrapper=PydanticConfiguration,
                wrapped=configuration_class,
                assigned=missing_in__dict__ + list(WRAPPER_ASSIGNMENTS),
                updated=(),  # default of '__dict__' does not apply here, see missing_in__dict__
            )

            cls._register_application(algorithm_class, PydanticConfiguration)

            return PydanticConfiguration

        return decorator

    @staticmethod
    def configuration_class_as_tuple(
        algorithm_configuration_class: Type[AlgorithmConfiguration],
    ) -> "ConfigurationTuple":
        """Get a hashable identifier per application."""
        return ConfigurationTuple(
            algorithm_type=algorithm_configuration_class.algorithm_type,
            domain=algorithm_configuration_class.domain,
            algorithm_name=algorithm_configuration_class.algorithm_name,
            algorithm_application=algorithm_configuration_class.algorithm_application,
        )

    @classmethod
    def get_application(
        cls,
        algorithm_type: str,
        domain: str,
        algorithm_name: str,
        algorithm_application: str,
    ) -> AlgorithmApplication:
        return cls.applications[
            ConfigurationTuple(
                algorithm_type=algorithm_type,
                domain=domain,
                algorithm_name=algorithm_name,
                algorithm_application=algorithm_application,
            )
        ]

    @classmethod
    def get_matching_configuration_defaults(
        cls,
        algorithm_type: str,
        domain: str,
        algorithm_name: str,
        algorithm_application: str,
    ) -> Dict[str, AnnotationTuple]:
        Configuration = cls.get_application(
            algorithm_type=algorithm_type,
            domain=domain,
            algorithm_name=algorithm_name,
            algorithm_application=algorithm_application,
        ).configuration_class

        defaults_dict = {}
        for (
            argument,
            default_value,
        ) in Configuration().__dict__.items():
            defaults_dict[argument] = AnnotationTuple(
                annotation=Configuration.__annotations__[argument],
                default_value=default_value,
            )
        return defaults_dict

    @classmethod
    def get_matching_configuration_schema(
        cls,
        algorithm_type: str,
        domain: str,
        algorithm_name: str,
        algorithm_application: str,
    ) -> Dict[str, Any]:
        Configuration = cls.get_application(
            algorithm_type=algorithm_type,
            domain=domain,
            algorithm_name=algorithm_name,
            algorithm_application=algorithm_application,
        ).configuration_class
        return Configuration.__pydantic_model__.schema()  # type: ignore

    @classmethod
    def get_configuration_instance(
        cls,
        algorithm_type: str,
        domain: str,
        algorithm_name: str,
        algorithm_application: str,
        *args,
        **kwargs,
    ) -> AlgorithmConfiguration:
        """Create an instance of the matching AlgorithmConfiguration from the ApplicationsRegistry.

        Args:
            algorithm_type: general type of generative algorithm.
            domain:  general application domain. Hints at input/output types.
            algorithm_name: name of the algorithm to use with this configuration.
            algorithm_application: unique name for the application that is the use of this
                configuration together with a specific algorithm.
            algorithm_version: to differentiate between different versions of an application.
            *args: additional positional arguments passed to the configuration.
            **kwargs: additional keyword arguments passed to the configuration.

        Returns:
            an instance of the configuration.
        """
        Configuration = cls.get_application(
            algorithm_type=algorithm_type,
            domain=domain,
            algorithm_name=algorithm_name,
            algorithm_application=algorithm_application,
        ).configuration_class
        return Configuration(*args, **kwargs)

    @classmethod
    def get_application_instance(
        cls,
        algorithm_type: str,
        domain: str,
        algorithm_name: str,
        algorithm_application: str,
        target: Any = None,
        **kwargs,
    ) -> GeneratorAlgorithm:
        """Instantiate an algorithm via a matching application from the ApplicationsRegistry.

        Additional arguments are passed to the configuration and override any arguments
        in the ApplicationsRegistry.

        Args:
            algorithm_type: general type of generative algorithm.
            domain:  general application domain. Hints at input/output types.
            algorithm_name: name of the algorithm to use with this configuration.
            algorithm_application: unique name for the application that is the use of this
                configuration together with a specific algorithm.
            algorithm_version: to differentiate between different versions of an application.
            target: optional context or condition for the generation.
            **kwargs: additional keyword arguments passed to the configuration.

        Returns:
            an instance of a generative algorithm ready to sample from.
        """
        application_tuple = cls.get_application(
            algorithm_type=algorithm_type,
            domain=domain,
            algorithm_name=algorithm_name,
            algorithm_application=algorithm_application,
        )
        parameters = {
            key: annotation_tuple.default_value
            for key, annotation_tuple in application_tuple.parameters_dict.items()
        }
        parameters.update(kwargs)

        return application_tuple.algorithm_class(
            configuration=application_tuple.configuration_class(**parameters),
            target=target,
        )

    @classmethod
    def list_available(cls) -> List[Dict[str, str]]:
        available = []
        for config_tuple, application in cls.applications.items():
            available.extend(
                [
                    dict(**config_tuple._asdict(), algorithm_version=version)
                    for version in application.configuration_class.list_versions()
                ]
            )
        return available



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from .graph_building_env import GraphBuildingEnv, GraphBuildingEnvContext
from .mol_building_env import MolBuildingEnvContext


# convert to a factory/registry
def build_env_context(
    environment_name: str = "graph_building_env",
    context_name: str = "graph_building_env_context",
):

    """Builds an environment and context environment.

    Args:
        environment_name: The name of the environment to build.
        context_name: The name of the context environment to build.

    Returns:
        tuple with selected environment and context environment.
    """

    env, context = None, None
    if environment_name == "graph_building_env":
        env = GraphBuildingEnv

    if context_name == "graph_building_env_context":
        context = GraphBuildingEnvContext
    elif context_name == "mol_building_env_context":
        context = MolBuildingEnvContext

    return env, context
